import numpy as np
import netCDF4 as nc

import yori.gridtools as grd

from datetime import datetime
from pkg_resources import get_distribution


NC_FILL_DOUBLE = -9999.
VERSION = get_distribution('yori').version


def daily_mask(satellite, daily, flist, gridsize):
    if satellite == '':
        raise ValueError('--daily requires the --method option')
    daily_flist = []
    try:
        indate = datetime.strptime(daily, '%Y-%m-%d')
    except ValueError:
        raise ValueError('Date not recognized. The prescribed format is: ' +
                         'YYYY-MM-DD')

    for fname in flist:
        f = nc.Dataset(fname)
        try:
            timeattr = f.getncattr('time_coverage_start')
            ftime = parse_time_attr(timeattr)
        except AttributeError:
            raise AttributeError('Global attribute "time_coverage_start" ' +
                                 'not found in input file ' + fname)

        diff_time = ftime - indate
        if np.abs(diff_time.days) > 1:
            raise IOError("Input date and file date don't match. " +
                          "Time difference is greater than one day")

        if np.abs(diff_time.days) == 0:
            daily_flist.append(fname)
        elif ((diff_time.days == -1) and
              (np.abs(diff_time.total_seconds()/60.) < 181)):
            daily_flist.append(fname)
        elif ((diff_time.days == 1) and
              (np.abs((diff_time.total_seconds()-86400.)/60.) < 180)):
            daily_flist.append(fname)
        f.close()

    flist = daily_flist
    if satellite == 'c6aqua':
        dmask = aqua_daily_aggr(timeattr, daily, gridsize)
    if satellite == 'c6terra':
        dmask = terra_daily_aggr(timeattr, daily, gridsize)

    return dmask


def prepare_vars(var_add, group_batch, ncfile, granule_edges):
    for grp in group_batch:
        var_add[grp] = {}
        # define a bounding box to limit I/O by scanning Pixel_Counts
        # for nonzero entries
#        if 'min_pixel_counts' in grp:
#            var_add[grp]['min_pixel_counts'] = ncfile[grp]['min_pixel_counts']

        if 'Pixel_Counts' in ncfile[grp].variables:
            cnt = ncfile[grp]['Pixel_Counts'][:]
            x, y = cnt.nonzero()
            if len(x) == 0:
                del var_add[grp]
                continue
            x_slice = slice(x.min(), x.max() + 1)
            y_slice = slice(y.min(), y.max() + 1)
            var_add[grp]['Pixel_Counts'] = cnt[x_slice, y_slice]
            var_add[grp]['_slices'] = (x_slice, y_slice)
            var_add[grp]['Num_Days'] = compute_valid_days(cnt)[x_slice, y_slice]
        elif granule_edges:
            x_slice, y_slice = granule_edges
            var_add[grp]['_slices'] = (x_slice, y_slice)
        else:
            x_slice = slice(None)  # gotta read the whole grid...
            y_slice = slice(None)

        for w in ncfile[grp].variables:
            if w != ['Pixel_Counts', 'min_pixel_counts', 'min_valid_days']:
                if w != ['Median_Distribution']:
                    var_add[grp][w] = ncfile[grp][w][x_slice, y_slice]
#                else:
#                    var_add['edges'] = ncfile[grp][w].Median_Bins

    return var_add


def initialize_arrays(group_batch, ncfile, latkey, lonkey, fv):
    var_in = {}
    for grp in group_batch:
        if grp == latkey or grp == lonkey:
            pass
        else:
            var_in[grp] = {}
            if hasattr(ncfile[grp], 'min_pixel_counts'):
                var_in[grp]['min_pixel_counts'] = ncfile[grp].min_pixel_counts
            if hasattr(ncfile[grp], 'min_valid_days'):
                var_in[grp]['min_valid_days'] = ncfile[grp].min_valid_days
                var_in[grp]['Num_Days'] = ncfile[grp]['Pixel_Counts'][:] * 0.
            for v in ncfile[grp].variables:
                if hasattr(ncfile[grp][v], 'Median_Bins'):
                    var_in[grp]['edges'] = ncfile[grp][v].Median_Bins
                var_in[grp][v] = ncfile[grp][v][:] * 0.
        # multipled by 0 so that the first var_in doesn't need to be masked
        # for the daily aggregation. This wouldn't be necessary for non-daily
        # aggregations

    # zero out fill values now, instead of at each loop iteration
    # perhaps even better would be to not use a fill value, but rather
    # user zero as the default value for all arrays that simply sum up
    # results (everything except mean and stdev)
    for group_vals in var_in.values():
        for var, arr in group_vals.items():
            if var not in ['min_pixel_counts', 'min_valid_days']:
                arr = arr.view(np.ndarray)  # converts from masked array without copying
                arr[arr == fv] = 0
                group_vals[var] = arr

    return var_in


def gq_aggregate(curr_vals, new_vals, daily_mask, fill_value):
    """This function is an attempt to speed up the aggregation logic

    Mimics yori.aggregator.aggregator. Unlike that function, the return value
    contains the same arrays as are passed in via curr_vals, not copies.
    """
    slices = new_vals.pop('_slices', (slice(None), slice(None)))
    daily_mask = np.asarray(daily_mask, bool)[slices]
    derived_vars = ['Mean', 'Standard_Deviation', 'Median',
                    'Weighted_Mean', 'Weighted_Standard_Deviation',
                    'min_pixel_counts', 'min_valid_days', 'edges',]
    for var in curr_vals:
        if var in derived_vars:
            continue
        curr = curr_vals[var][slices]
        new = new_vals[var].view(np.ndarray)
        new[new == fill_value] = 0
        if var == 'Min':
            # need to turn zeros into NaNs to get the min, otherwise every time the min()
            # will always pick the zero, while that should be a value to not take into
            # account.
            curr[curr == 0] = np.nan
            new[new == 0] = np.nan
            curr[daily_mask] = np.nanmin([curr, new], axis=0)[daily_mask]
        elif var == 'Max':
            curr[daily_mask] = np.nanmax([curr, new], axis=0)[daily_mask]
        else:
            curr[daily_mask] += new[daily_mask]

    return curr_vals


def compute_valid_days(pixel_counts):
    valid_days = np.zeros(pixel_counts.shape)
    valid_days[pixel_counts != 0] = 1
    return valid_days


def compute_aggr_stats(var_in, fv):
    for vals in var_in.values():
        if 'Pixel_Counts' not in vals:
            continue
        vals['Mean'] = vals['Sum'] / vals['Pixel_Counts']
        vals['Standard_Deviation'] = np.sqrt(
            vals['Sum_Squares'] / vals['Pixel_Counts']
            - (vals['Sum'] * vals['Sum']) / (vals['Pixel_Counts'] * vals['Pixel_Counts']))

        if 'Median' in vals:
            edges = vals['edges']
            resolution = np.diff(vals['edges'])[0]
            vals['Median'] = aggregate_median(vals['Median_Distribution'], edges,
                                              resolution)

        if 'Weights' in vals:
            vals['Weighted_Mean'] = vals['Weighted_Sum'] / vals['Weights']
            vals['Weighted_Standard_Deviation'] = np.sqrt(
                    vals['Weighted_Sum_Squares'] / vals['Weights'] -
                    (vals['Weighted_Sum'] / vals['Weights'])**2)

        '''
        if 'min_pixel_counts' in vals:
            vals['min_pixel_counts'] = vals['min_pixel_counts']
        if 'min_valid_days' in vals:
            vals['min_valid_days'] = vals['min_valid_days']
        '''

        no_data = (vals['Pixel_Counts'] == 0)

        for k in list(vals):
            if k in ['Pixel_Counts', 'Weights', 'min_pixel_counts', 'min_valid_days',
                     'edges']:
                continue
            else:
                vals[k][no_data] = fv
                # vals[k][np.isnan(vals[k])] = fv


def restructDict(temp_var):
    tmp_out_var = {}
    for k in list(temp_var):
        if type(temp_var[k]) is dict:
            for kk in list(temp_var[k]):
                if k[-1] == '/':
                    k_short = k[:-1]
                else:
                    k_short = k
                tmp_out_var[k_short + '/' + kk] = temp_var[k][kk]
        else:
            tmp_out_var[k] = temp_var[k]

    return tmp_out_var


def aggregate_median(median_dist, edges, res):
    dims = median_dist.shape
    var = np.reshape(median_dist, (dims[0]*dims[1], dims[2]))
    aggr_median = np.zeros((dims[0]*dims[1]))
    for i in range(dims[0]*dims[1]):
        if np.sum(var[i, :]) > 0.0:
            aggr_median[i] = grd.estimate_median(var[i, :], edges, res)['est_median']

    return aggr_median.reshape(dims[0], dims[1])


########################################
#               FUNCTION               #
########################################
#
def parse_time_attr(value):
    for fmt in [
         '%Y-%m-%dT%H:%M:%SZ',
         '%Y-%m-%dT%H:%M:%S.%fZ',
         '%Y-%m-%dT%H:%M:%S.%f']:
        try:
            return datetime.strptime(value, fmt)
        except ValueError:
            pass

    raise ValueError('could not parse time attr', value)


########################################
#               FUNCTION               #
########################################
#
def aqua_daily_aggr(timeattr, daily, gridsize):

    ftime = parse_time_attr(timeattr)
    indate = datetime.strptime(daily, '%Y-%m-%d')

    diff_time = ftime - indate

    # -180 to 180
    if (diff_time.days == 0) and (ftime.hour >= 3.) and (ftime.hour < 24.):
        coords_idx = np.array(range(0, int(360/gridsize)))

    # -90 to 0 and 90 to 180
    elif (diff_time.days == 0) and (ftime.hour < 3.):
        coords_idx = np.array(range(int(90/gridsize), int(180/gridsize)))
        coords_idx = np.append(coords_idx, range(int(270/gridsize),
                                                 int(360/gridsize)))

    # -180 to -90 and 0 to 90
    elif (diff_time.days == 1) and (ftime.hour < 3.):
        coords_idx = np.array(range(0, int(90/gridsize)))
        coords_idx = np.append(coords_idx, range(int(180/gridsize),
                                                 int(270/gridsize)))

    else:
        raise IOError('The hour in the time_coverage_start attribute ' +
                      'is out of range')

    daily_mask = np.zeros((int(360/gridsize), int(180/gridsize)))
    daily_mask[coords_idx, :] = 1

    return np.array(daily_mask)


def terra_daily_aggr(timeattr, daily, gridsize):

    ftime = parse_time_attr(timeattr)
    indate = datetime.strptime(daily, '%Y-%m-%d')

    diff_time = ftime - indate

    # -180 to 180
    if (diff_time.days == 0) and (ftime.hour >= 0.) and (ftime.hour < 21.):
        coords_idx = np.array(range(0, int(360/gridsize)))

    # -180 to -90 and 0 to 90
    elif (diff_time.days == 0) and (ftime.hour >= 21.):
        coords_idx = np.array(range(int(0/gridsize), int(90/gridsize)))
        coords_idx = np.append(coords_idx, range(int(180/gridsize),
                                                 int(270/gridsize)))

    # -90 to 0 and 90 to 180
    elif (diff_time.days == -1) and (ftime.hour >= 21.):
        coords_idx = np.array(range(int(90/gridsize), int(180/gridsize)))
        coords_idx = np.append(coords_idx, range(int(270/gridsize),
                                                 int(360/gridsize)))

    else:
        raise IOError('The hour in the time_coverage_start attribute ' +
                      'is out of range')


#   .--.      .-'.      .--.      .--.      .--.      .--.      .`-.      .--.
# :::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\
# '      `--'      `.-'      `--'      `--'      `--'      `-.'      `--'
#
#     # This is the MODIS implementation of the daily aggregation
#
#     # -180 to 180
#     if (diff_time.days == 0) and (ftime.hour >= 3.) and (ftime.hour <= 24.):
#         coords_idx = np.array(range(0, int(360/gridsize)))
#
#     # -180 to -90 and 0 to 90
# #     elif (diff_time.days == 0) and (ftime.hour >= 21.):
# #         coords_idx = np.array(range(int(0/gridsize), int(90/gridsize)))
# #         coords_idx = np.append(coords_idx, range(int(180/gridsize),
# #                                                  int(270/gridsize)))
#
#     # -90 to 0 and 90 to 180
#     elif (diff_time.days == 0) and (ftime.hour < 3.):
#         coords_idx = np.array(range(int(90/gridsize), int(180/gridsize)))
#         coords_idx = np.append(coords_idx, range(int(270/gridsize),
#                                                 int(360/gridsize)))
#
#     # -180 to -90 and 0 to 90
#     elif (diff_time.days == 1) and (ftime.hour < 3.):
#         coords_idx = np.array(range(0, int(90/gridsize)))
#         coords_idx = np.append(coords_idx, range(int(180/gridsize),
#                                                  int(270/gridsize)))
#
#     # -90 to 0 and 90 to 180
# #     elif (diff_time.days == -1) and (ftime.hour >= 21.):
# #         coords_idx = np.array(range(int(90/gridsize), int(180/gridsize)))
# #         coords_idx = np.append(coords_idx, range(int(270/gridsize),
# #                                                  int(360/gridsize)))
#
#     else:
#         raise IOError('The hour in the time_coverage_start attribute' +
#                       'is out of range')
#
#   .--.      .-'.      .--.      .--.      .--.      .--.      .`-.      .--.
# :::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\::::::::.\
# '      `--'      `.-'      `--'      `--'      `--'      `-.'      `--'

    daily_mask = np.zeros((int(360/gridsize), int(180/gridsize)))
    daily_mask[coords_idx, :] = 1

    # probably check if all the files are provided (30 hours)
    # use datetime.datetime.strptime()
    # date string format '%Y-%m-%dT%H:%M:%S.000Z'

    # Longitude intervals:
    #     0   - 90   21:00 - 21:00
    #     90  - 180  03:00 - 03:00
    #     180 - 270  21:00 - 21:00
    #     270 - 360  03:00 - 03:00

    # Longitude intervals:
    #     0   - 90   03:00 - 03:00
    #     90  - 180  21:00 - 21:00
    #     180 - 270  03:00 - 03:00
    #     270 - 360  21:00 - 21:00

    return np.array(daily_mask)



import os
from subprocess import run
from netCDF4 import Dataset

import numpy as np

import pytest


@pytest.fixture()
def fixturepath():
    return os.path.join(os.path.dirname(__file__), '../fixtures')


# GRIDDING TESTS
# config files
@pytest.fixture()
def dummy_config(fixturepath):
    return os.path.join(fixturepath, 'testcfg.yml')


@pytest.fixture()
def realcase_config(fixturepath):
    return os.path.join(fixturepath, 'fulltest_config.yml')


# input files
@pytest.fixture()
def dummy_inputfile(tmpdir, fixturepath):
    run('ncgen -b -o {0}/testinput.nc {1}/test.cdl'.format(tmpdir, fixturepath),
        shell=True, check=True)
    return os.path.join(str(tmpdir), 'testinput.nc')


@pytest.fixture()
def realcase_input(fixturepath):
    return os.path.join(str(fixturepath), 'preyori_ref_testfile_1.nc')


# output files
@pytest.fixture()
def dummy_outputfile(tmpdir):
    return os.path.join(str(tmpdir), 'testoutput.nc')


@pytest.fixture()
def realcase_ref_output(fixturepath):
    return os.path.join(str(fixturepath), 'grid_ref_testfile_1.nc')


@pytest.fixture()
def realcase_output(tmpdir, fixturepath):
    return os.path.join(str(tmpdir), 'grid_out_testfile_1.nc')


# run gridding tests
def test_grid_dummy(dummy_config, dummy_inputfile, dummy_outputfile):
    run('yori-grid {0} {1} {2} -c 5'.format(dummy_config, dummy_inputfile,
                                            dummy_outputfile),
        shell=True, check=True)

    assert os.path.exists(dummy_outputfile)
    assert os.stat(dummy_outputfile).st_size != 0


def test_grid_realcase(realcase_config, realcase_input, realcase_ref_output,
                       realcase_output):
    run('yori-grid {0} {1} {2} -c 5'.format(realcase_config, realcase_input,
                                            realcase_output),
        shell=True, check=True)
    assert os.path.exists(realcase_output)

    check_differences(realcase_ref_output, realcase_output)


# AGGREGATION TEST
# input files
@pytest.fixture()
def aggr_filelist(tmpdir, fixturepath):
    run('ncgen -b -o {0}/grid_input_1.nc {1}/test_aggr_1.cdl'.format(tmpdir, fixturepath),
        shell=True, check=True)
    run('ncgen -b -o {0}/grid_input_2.nc {1}/test_aggr_2.cdl'.format(tmpdir, fixturepath),
        shell=True, check=True)
    run('ls -1 {0}/grid_input_?.nc > {0}/aggr_list'.format(tmpdir),
        shell=True, check=True)
    return os.path.join(str(tmpdir), 'aggr_list')


# filelists for the aggregation tests
@pytest.fixture()
def aggr_realcase_filelist(tmpdir, fixturepath):
    run('ls -1 {0}/grid_ref_testfile_?.nc > {1}/aggr_realcase_list'.format(fixturepath,
                                                                           tmpdir),
        shell=True, check=True)
    return os.path.join(str(tmpdir), 'aggr_realcase_list')


@pytest.fixture()
def aggr_daily_filelist(tmpdir, fixturepath):
    run('ls -1 {0}/grid_ref_testfile_?.nc > {1}/aggr_daily_list'.format(fixturepath,
                                                                        tmpdir),
        shell=True, check=True)
    return os.path.join(str(tmpdir), 'aggr_daily_list')


# output files
@pytest.fixture()
def dummy_aggr_outputfile(tmpdir):
    return os.path.join(str(tmpdir), 'aggr_testoutput.nc')


@pytest.fixture()
def realcase_aggr_output(tmpdir):
    return os.path.join(str(tmpdir), 'aggr_test_realcase.nc')


@pytest.fixture()
def realcase_daily_output(tmpdir):
    return os.path.join(str(tmpdir), 'aggr_daily_realcase.nc')


@pytest.fixture()
def realcase_ref_aggr_output(fixturepath):
    return os.path.join(str(fixturepath), 'aggr_ref_realcase.nc')


# run aggregation tests
def test_aggr(aggr_filelist, dummy_aggr_outputfile):
    run('yori-aggr {0} {1} -c 5'.format(aggr_filelist, dummy_aggr_outputfile),
        shell=True, check=True)

    assert os.path.exists(dummy_aggr_outputfile)
    assert os.stat(dummy_aggr_outputfile).st_size != 0


def test_aggr_realcase(aggr_realcase_filelist, realcase_aggr_output,
                       realcase_ref_aggr_output):
    run('yori-aggr {0} {1} -c 5'.format(aggr_realcase_filelist,
                                        realcase_aggr_output),
        shell=True, check=True)

    assert os.path.exists(realcase_aggr_output)

    check_differences(realcase_ref_aggr_output, realcase_aggr_output)


def test_daily_aggr(aggr_daily_filelist, realcase_daily_output):
    methods = ['c6aqua', 'c6terra']
    for method in methods:
        run(('yori-aggr {0} {1} -c 5 --daily 2014-02-02 ' +
             '--method {2}').format(aggr_daily_filelist, realcase_daily_output,
                                    method),
            shell=True, check=True)

        assert os.path.exists(realcase_daily_output)


def check_differences(ref_output, test_output):
    ref_file = Dataset(ref_output)
    new_file = Dataset(test_output)

    for var in list(ref_file.variables):
        check = np.allclose(ref_file[var][:].filled(ref_file[var]._FillValue),
                            new_file[var][:].filled(new_file[var]._FillValue),
                            equal_nan=True)
        assert check, 'Test failed at "{}"'.format(var)
    for grp in list(ref_file.groups):
        for var in list(ref_file[grp].variables):
            check = np.allclose(ref_file[grp][var][:].filled(ref_file[grp][var]._FillValue),
                                new_file[grp][var][:].filled(new_file[grp][var]._FillValue),
                                equal_nan=True)
            assert check, 'Test failed at "{}, {}"'.format(grp, var)

    ref_file.close()
    new_file.close()



import sys
import os.path

# from ..yori_aggregate import aggregate
from ..yori_merge import merge
from pkg_resources import get_distribution

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Command to two gridded files ' +
                                     'produced with either yori-grid or yori-aggr')
    parser.add_argument('-v', '--version', action='version', version=VERSION,
                        help='Show version and exit')
    parser.add_argument('filelist',
                        help='Can be a file path to an input file containing a single ' +
                        'input file per line, or "-" to read input list one per line ' +
                        'from standard input, or a comma separated list of files to ' +
                        'use as input.')
    parser.add_argument('output',
                        help='output file name and path. If --replace is '
                             'used, file1 is renamed instead.')
#    parser.add_argument('-r', '--replace',
#                        help='merge does not create a new merged file, instead, it add '
#                             'variables from file2 into file1. If this option is used '
#                             'the argument "output" renames file1', action='store_true')
    parser.add_argument('-c', '--compression', default=5,
                        help='set the compression level for the output file (valid '
                             'range 1-9)')
    parser.add_argument('-f', '--fill-value',
                        help='allows the user to overwrite the fill values in the '
                             'input files')

    args = parser.parse_args()

    # read file list from stdin
    if args.filelist == '-':
        inputs = [l.strip() for l in sys.stdin.read().split('\n') if l.strip()]

    # if the file exists, assume it is a file and try to fetch inputs
    elif os.path.exists(args.filelist):
        inputs = [l.strip() for l in open(args.filelist).read().split('\n') if l.strip()]

    else:
        inputs = [f.strip() for f in args.filelist.split(',')]

    merge(inputs, args.output,
          compression=args.compression,
          fill_value=args.fill_value)



try:
    from ruamel import yaml
except ImportError:
    import ruamel_yaml as yaml

# import yaml
import numpy as np


########################################
#               FUNCTION               #
########################################
# function to read the configuration file
class ConfigReader(object):

    #
    def __init__(self, yaml_str):

        yaml_str = str(yaml_str[:])
        # read config
        # config = yaml.load(yaml_str, yaml.RoundTripLoader)
        config = yaml.safe_load(yaml_str)

        self.grid_settings = config['grid_settings']
        self.var_settings = config['variable_settings']
        self.yaml_str = yaml_str

        # read the YAML file and check some exceptions
        var_list_in = []
        var_list_out = []

        for i in range(len(self.var_settings)):
            if any(names not in self.var_settings[i]
                   for names in ['name_in', 'name_out']):
                raise IOError('in YAML file "name_in" or "name_out" ' +
                              'are missing somewhere in the file')
            var_list_in.append(self.var_settings[i]['name_in'])
            if 'masks' in self.var_settings[i]:
                try:
                    len(self.var_settings[i]['masks'])
                except TypeError:
                    raise IOError('in YAML file "masks" keyword has been ' +
                                  'defined without any mask names')
            if 'inverse_masks' in self.var_settings[i]:
                try:
                    len(self.var_settings[i]['inverse_masks'])
                except TypeError:
                    raise IOError('in YAML file "inverse_masks" keyword has ' +
                                  'been defined without any mask names')
            var_list_out.append(self.var_settings[i]['name_out'])

            if '2D_histograms' in self.var_settings[i]:
                # if 'histograms' not in self.var_settings[i]:
                #     raise IOError('in YAML file, "histograms" is not ' +
                #                   'defined. In order to compute a joint' +
                #                   '(2D) histogram the "histograms" ' +
                #                   'keyword must be defined for the first ' +
                #                   'variable')
                jhisto = self.var_settings[i]['2D_histograms']
                for j in range(len(jhisto)):
                    if 'name_out' not in jhisto[j]:
                        raise IOError('in YAML file, "name_out" is missing ' +
                                      'from some 2D_histogram')
                    jvar = jhisto[j]['joint_var']
                    if 'name_in' not in jvar:
                        raise IOError('in YAML file, "name_in" is missing ' +
                                      'from some 2D_histogram')
                    if 'extra_masks' in jvar:
                        try:
                            len(jvar['extra_masks'])
                        except TypeError:
                            raise IOError('in YAML file "extra_masks" ' +
                                          'keyword has been defined without ' +
                                          'any mask names')

        if any(var_list_out.count(x) > 1 for x in var_list_out):
            raise NameError('in YAML file one or more output variables have ' +
                            'the same name')

    def read_variable_list(self, var_block):
        var_list = {'in': [], 'out': [], 'masks': [], 'inverse_masks': [],
                    'histograms': [], '2D_histograms': [], 'extra_masks': [],
                    'only_histogram': []}

        for v in self.var_settings:
            if any(names not in v for names in ['name_in', 'name_out']):
                raise IOError('in YAML file "name_in" or "name_out" ' +
                              'are missing somewhere in the file')
            var_list['in'].append(v['name_in'])
            var_list['out'].append(v['name_out'])

        return var_list

    def readHist(self, histBlock):

        if 'edges' in histBlock:
            varbins = list(histBlock['edges'])

        elif all(key in histBlock for key in ['start', 'stop', 'interval']):
            varbins = list(np.arange(histBlock['start'], histBlock['stop'],
                                     histBlock['interval']))
#        elif all(key in histBlock for key in ['resolution', 'min_value', 'max_value']):
#            varbins = list(np.arange(histBlock['min_value'], histBlock['max_value'],
#                                     histBlock['resolution']))
        else:
            raise IOError('in YAML file, histogram is not defined correctly')

        return varbins



import numpy as np

from pkg_resources import get_distribution

VERSION = get_distribution('yori').version
_int_variables = ['Pixel_Counts', 'Num_Days']


def write_output(fout, gridvar, compression, grid_size, ymlSet, fv,
                 edges=[], write_edges=False):

    if 'latitude' not in fout.variables:
        fout.createDimension('latitude', 180/grid_size)
        fout.createVariable('latitude', 'f8', ('latitude', ),
                            fill_value=fv)
        fout['latitude'][:] = gridvar['latitude']
        setattr(fout.variables['latitude'], 'units', 'degrees_north')
    if 'longitude' not in fout.variables:
        fout.createDimension('longitude', 360/grid_size)
        fout.createVariable('longitude', 'f8', ('longitude', ),
                            fill_value=fv)
        fout['longitude'][:] = gridvar['longitude']
        setattr(fout.variables['longitude'], 'units', 'degrees_east')

    for k in list(gridvar):
        if k == 'latitude' or k == 'longitude':
            continue
        if k.split('/')[-1] in _int_variables or gridvar[k].ndim > 2:
            vartype = 'i4'
        else:
            vartype = 'f8'

        dims = create_dims(fout, gridvar, k)
        fout.createVariable(k, vartype, dims,
                            complevel=int(compression), zlib=True,
                            fill_value=fv)
        if edges == []:
            fout[k][:] = gridvar[k]
        else:
            fout[k][edges[0]:edges[1]+1, edges[2]:edges[3]+1] = (
                gridvar[k][edges[0]:edges[1]+1, edges[2]:edges[3]+1])

    if write_edges:
        fout.granule_edges = np.array(edges, 'i4')


def create_dims(fout, variable, k):
    default_dims_name = ('longitude', 'latitude')
    dim_list = tuple(fout.dimensions.keys())
    dims = default_dims_name

    # histo_dim_name, jhisto_dim_name = '', ''
    histo_dim_name, jhisto_primary_name, jhisto_joint_name = '', '', ''
    if variable[k].ndim == 3:
        histo_dim_name = 'Histo_{0}_{1}'.format(k.split('/')[0],
                                                variable[k].shape[2]).lower()
    if variable[k].ndim == 4:
        primary_var = k.split('/')[0]
        joint_var = k.split('/')[-1].replace('JHisto_vs_', '')
        jhisto_primary_name = 'jhisto_{0}_{1}'.format(primary_var,
                                                      variable[k].shape[2]).lower()
        jhisto_joint_name = 'jhisto_{0}_{1}'.format(joint_var,
                                                    variable[k].shape[3]).lower()
        # jhisto_dim_name = 'JHisto_{0}_{1}'.format(k.split('/')[-1],
        #                                           variable[k].shape[3]).lower()

    if histo_dim_name not in dim_list and variable[k].ndim == 3:
        fout.createDimension(histo_dim_name, variable[k].shape[2])
    if jhisto_primary_name not in dim_list and variable[k].ndim == 4:
        fout.createDimension(jhisto_primary_name, variable[k].shape[2])
    if jhisto_joint_name not in dim_list and variable[k].ndim == 4:
        fout.createDimension(jhisto_joint_name, variable[k].shape[3])
        # fout.createDimension(jhisto_dim_name, variable[k].shape[3])

    if histo_dim_name != '':
        dims += (histo_dim_name, )
    if jhisto_primary_name != '':
        dims += (jhisto_primary_name, )
    if jhisto_joint_name != '':
        dims += (jhisto_joint_name, )

    return dims


def append_attributes(fout, var_settings):
    grp = fout.groups[var_settings['name_out']]
    for var in list(grp.variables):
        setattr(grp[var], 'title', var_settings['name_out'] + ': ' + var)

    for attr in var_settings.get('attributes', []):
        setattr(grp, attr['name'], attr['value'])

        if (attr['name'] == 'units'):
            if 'Mean' in list(grp.variables):
                setattr(grp['Mean'], attr['name'], attr['value'])
            if 'Standard_Deviation' in list(grp.variables):
                setattr(grp['Standard_Deviation'], attr['name'], attr['value'])

    if 'histograms' in var_settings:
        if 'edges' in var_settings['histograms']:
            hist_bins = var_settings['histograms']['edges']
            setattr(grp['Histogram_Counts'], 'Histogram_Bin_Boundaries',
                    np.array(hist_bins, dtype='f8'))
        else:
            setattr(grp['Histogram_Counts'], 'Histogram_Bin_Boundaries',
                    compute_hist_bins(var_settings['histograms']))

    if 'median' in var_settings:
        setattr(grp['Median_Distribution'], 'Median_Bins',
                    compute_hist_bins(var_settings['median']))

    if '2D_histograms' in var_settings:
        for hist2d in var_settings['2D_histograms']:
            if 'edges' in hist2d['primary_var']:
                setattr(grp[hist2d['name_out']], 'JHisto_Bin_Boundaries',
                        np.array(hist2d['primary_var']['edges'], 'f8'))
            else:
                setattr(grp[hist2d['name_out']], 'JHisto_Bin_Boundaries',
                        compute_hist_bins(hist2d['primary_var']))
            if 'edges' in hist2d['joint_var']:
                setattr(grp[hist2d['name_out']],
                        'JHisto_Bin_Boundaries_Joint_Parameter',
                        np.array(hist2d['joint_var']['edges'], 'f8'))
            else:
                setattr(grp[hist2d['name_out']],
                        'JHisto_Bin_Boundaries_Joint_Parameter',
                        compute_hist_bins(hist2d['joint_var']))

    if 'min_pixel_counts' in var_settings:
        setattr(grp, 'min_pixel_counts', var_settings['min_pixel_counts'])
    if 'min_valid_days' in var_settings:
        setattr(grp, 'min_valid_days', var_settings['min_valid_days'])


def compute_hist_bins(hist_data):
    hist_bins = np.arange(hist_data['start'], hist_data['stop'], hist_data['interval'],
                          dtype='f8')
    return hist_bins



import numpy as np

from pyhdf.SD import SD


########################################
#               FUNCTION               #
########################################
# function for reading variables inside an hdf4 file
def read_hdf(fstr, varlist=[]):
    """read_hdf(fstr[, varlist])

    read the variables provided as inputs from an hdf4 file and save them into a dictionary

    Parameters
    ----------
    fstr : string
        name of the hdf4file
    varlist : array [string]
        name of the variables to be read inside the file
        default = 0 (reads the entire file)

    Returns
    -------
    output : dictionary
        dictionary of variables contained in the input hdf4 file
    """
    #read open the hdf file
    hdff = SD(fstr);

    #define the string of variables to read if not defined
    if varlist != []:
        varlist = varlist;
    else:
        varlist = hdff.datasets();
        varlist = varlist.keys();
    #end var list files

    #define the dictionary
    output = {};

    for varname in varlist:
        vartmp    = hdff.select(varname);
        attr      = vartmp.attributes();
        attr_list = attr.keys();
        var       = vartmp.get();
        varf      = var.astype('float64')

        if '_FillValue' in attr_list:
            fillvalue = attr['_FillValue'];
            fillidx = np.where(varf == fillvalue);
            varf[fillidx]=0 #np.nan;
        else:
            fillvalue = 0
        if 'add_offset' in attr_list:
            add_offset = attr['add_offset'];
        else:
            add_offset = 0;
        if 'scale_factor' in attr_list:
            scale_factor = attr['scale_factor'];
        else:
            scale_factor = 1

        output[varname] = varf*scale_factor+add_offset;

    hdff.end()

    return output



import sys
import os.path

from ..yori_aggregate import aggregate
from pkg_resources import get_distribution

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Command to aggregate gridded files ' +
                                     'produced with yori-grid')
    parser.add_argument('--daily', metavar='YYYY-mm-dd', default='',
                        help='Indicate date for daily aggregation')
    parser.add_argument('--method',  default='',
                        help='Choose between C6 Aqua and Terra definition of day, ' +
                        '"c6aqua" and "c6terra" respectively')
    parser.add_argument('-V', '--verbose', action='store_true',
                        help='Enable verbose mode. Print on screen the list of files ' +
                        'while they are being processed')
    parser.add_argument('-v', '--version', action='version', version=VERSION,
                        help='Show version and exit')
    parser.add_argument('filelist',
                        help='Can be a file path to an input file containing a single ' +
                        'input file per line, or "-" to read input list one per line ' +
                        'from standard input, or a comma separated list of files to ' +
                        'use as input.')
    parser.add_argument('output', help='output file name and path')
    parser.add_argument('-f', '--force', action='store_true',
                        help='Force yori-aggr to ignore the daily flag in the input ' +
                        'files. It should be used with caution')
    parser.add_argument('-c', '--compression', default=5,
                        help='set the compression level for the output file (valid ' +
                        'range 1-9)')
    parser.add_argument('--min-pixel-counts', help='set the minimum number of pixels ' +
                        'for a grid cell to be included in the output D3 file',
                        action='store_true')
    parser.add_argument('--min-valid-days', help='set the minimum number of days with ' +
                        'non zero pixel counts in a grid cell in order to be included ' +
                        'in the M3 product', action='store_true')
    parser.add_argument('-b', '--batch-size', help='set batch size for the aggregation ' +
                        'process. Larger batch size can increase speed but requires ' +
                        'more memory. Default is 2, it should be changed with caution.',
                        default=2)
    args = parser.parse_args()

    # read file list from stdin
    if args.filelist == '-':
        inputs = [l.strip() for l in sys.stdin.read().split('\n') if l.strip()]

    # if the file exists, assume it is a file and try to fetch inputs
    elif os.path.exists(args.filelist):
        inputs = [l.strip() for l in open(args.filelist).read().split('\n') if l.strip()]

    else:
        inputs = [f.strip() for f in args.filelist.split(',')]

    aggregate(inputs,
              args.output,
              verbose=args.verbose,
              daily=args.daily,
              satellite=args.method,
              force=args.force,
              compression=args.compression,
              use_min_pixel_counts=args.min_pixel_counts,
              use_min_valid_days=args.min_valid_days,
              batch_size=args.batch_size)



import netCDF4 as nc

import yori.config_reader as cfg
import yori.IOtools as IO
import yori.aggrtools as tools

import sys
# import warnings

from pkg_resources import get_distribution

VERSION = get_distribution('yori').version
_attr_list = ['Yori_version', 'input_files', 'daily_defn_of_day_adjustment']


def merge(file_list, fname_out, compression=5, fill_value=None):
    f_in = nc.Dataset(file_list[0])
    f_out = nc.Dataset(fname_out, 'w')
    f_out.setncattr('Yori_version', VERSION)
    f_out.setncattr('daily_defn_of_day_adjustment',
                    f_in.getncattr('daily_defn_of_day_adjustment'))
    f_out.setncattr('YAML_config', f_in.getncattr('YAML_config'))
    f_out.setncattr('input_files', '')
    cfg_data = f_out.getncattr('YAML_config')
    ymlSet = cfg.ConfigReader(cfg_data)
    if fill_value is None:
        fv = ymlSet.grid_settings['fill_value']
    else:
        fv = fill_value

    lat = f_in['latitude'][:].data
    f_out.createDimension('latitude', len(lat))
    f_out.createVariable('latitude', 'f8', ('latitude', ), fill_value=fv)
    f_out['latitude'][:] = lat
    lon = f_in['longitude'][:].data
    f_out.createDimension('longitude', len(lon))
    f_out.createVariable('longitude', 'f8', ('longitude', ), fill_value=fv)
    f_out['longitude'][:] = lon

    f_in.close()
    f_out.close()

    for input_file in file_list:
        merge_files(input_file, fname_out, compression, fill_value)


def merge_files(input_file, output_file, compression, fill_value):
    f_out = nc.Dataset(output_file, 'a')
    f_in = nc.Dataset(input_file)
    ymlSet = cfg.ConfigReader(f_in.getncattr('YAML_config'))
    grid_size = ymlSet.grid_settings['gridsize']

    merge_vars = {}
    for grp in list(f_in.groups):
        merge_vars[grp] = {}
        for var in list(f_in[grp].variables):
            merge_vars[grp][var] = f_in[grp][var][:]
        merge_vars = tools.restructDict(merge_vars)
        IO.write_output(f_out, merge_vars, compression, grid_size, ymlSet, fill_value)
        merge_vars = {}

    for i in range(len(ymlSet.var_settings)):
        var_settings = ymlSet.var_settings[i]
        IO.append_attributes(f_out, var_settings)

    update_global_attributes(f_in, f_out)

    f_out.close()
    f_in.close()


def update_global_attributes(f_in, f_out):
    str_config_in = f_in.getncattr('YAML_config').split('variable_settings:')[1]
    flist = f_out.getncattr('input_files') + f_in.getncattr('input_files')

    str_config = f_out.getncattr('YAML_config') + str_config_in

    f_out.setncattr('input_files', flist)
    f_out.setncattr('YAML_config', str_config)


'''
def merge(in_file_1, in_file_2, out_fname, replace=False, compression=5, fill_value=None):

    if replace is True:
        # f_merged = nc.Dataset(in_file_1, 'a')
        # f1 = f_merged
        warnings.warn('This feature is still in development, the code will ignore '
                      'user input and run with "--replace False".')
    else:
        f1 = nc.Dataset(in_file_1)
        f_merged = nc.Dataset(out_fname, 'w')
    f2 = nc.Dataset(in_file_2)

#    for attr in _attr_list:
#        try:
#            attr1 = f1.getncattr(attr)
#            attr2 = f2.getncattr(attr)
#        except AttributeError:
#            pass

    # Attribute checks
    if 'Yori_version' not in f1.ncattrs() or 'Yori_version' not in f2.ncattrs():
        warnings.warn('yori-merge could not find "Yori_version" in the input files.')
    if f1.getncattr('Yori_version') != f2.getncattr('Yori_version'):
        warnings.warn('Yori_version warning #1')
    if (f1.getncattr('Yori_version') != str(VERSION) or
       f2.getncattr('Yori_version') != str(VERSION)):
        warnings.warn('Yori_version warning #2')
    f_merged.setncattr('Yori_version', str(VERSION))

    fl1, fl2 = '', ''
    if 'input_files' in f1.ncattrs():
        fl1 = f1.getncattr('input_files')
    if 'input_files' in f2.ncattrs():
        fl2 = f2.getncattr('input_files')
    f_merged.setncattr('input_files', fl1 + ',' + fl2)

    if ('daily_defn_of_day_adjustment' in f1.ncattrs() and
       'daily_defn_of_day_adjustment' in f2.ncattrs()):
        if (f1.getncattr('daily_defn_of_day_adjustment') !=
           f2.getncattr('daily_defn_of_day_adjustment')):
            warnings.warn('"daily_defn_of_day_adjustment" warning #1')
    elif ('daily_defn_of_day_adjustment' in f1.ncattrs() or
          'daily_defn_of_day_adjustment' in f2.ncattrs()):
        warnings.warn('"daily_defn_of_day_adjustment" warning #2')
    else:
        pass
    f_merged.setncattr('daily_defn_of_day_adjustment', '')

    config_str_1 = f1.getncattr('YAML_config')[:]
    config_str_2 = f2.getncattr('YAML_config')[:]

    yml_cfg1 = cfg.ConfigReader(config_str_1)
    yml_cfg2 = cfg.ConfigReader(config_str_2)

    if yml_cfg1.grid_settings['gridsize'] != yml_cfg2.grid_settings['gridsize']:
        raise ValueError('Value of "gridsize" in file 1 does not match '
                         'the value in file 2')
    if yml_cfg1.grid_settings['projection'] != yml_cfg2.grid_settings['projection']:
        raise ValueError('"projection" setting in file 1 is different from '
                         'the one in file 2')
    if (yml_cfg1.grid_settings['fill_value'] != yml_cfg2.grid_settings['fill_value']
       and fill_value is not None):
        warnings.warn('"fill_value" is set to different in the two input files, '
                      'in this situation Yori chooses the value for the first file '
                      'passed. If this is not the desired choice, you can define a '
                      'new fill_value by using the option --fill-value provided '
                      'with the merge tool ("merge --help" for more info)')

'''


if __name__ == "__main__":
    merge(sys.argv[1], sys.argv[2], replace=False)



from ..cluster_tools import hdf2nc
from pkg_resources import get_distribution

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Command to convert hdf4 files into netcdf4')
    parser.add_argument('filename', help='name of the file to convert')
    parser.add_argument('--varlist', default=[],
                        help='list of variables to include in the output file')
    parser.add_argument('-v', '--version', action='version', version=VERSION,
                        help='Show version and exit')
    parser.add_argument('-r', '--replace', action='store_true',
                        help='delete the hdf4 file after the conversion')
    parser.add_argument('-p', '--preproc', action='store_true',
                        help='preprocessing of the hdf4 before the conversion to netcdf4')
    parser.add_argument('-c', '--compression', default=0,
                        help='set netcdf compression level')
    parser.add_argument('-o', '--output', default='',
                        help='choose the output directory')
    args = parser.parse_args()

    if args.varlist == []:
        vlist = []
    else:
        vlist = [v.strip() for v in args.varlist.split(',')]

    hdf2nc(args.filename,
           var_list=vlist,
           replace=args.replace,
           preprocessing=args.preproc,
           compression=int(args.compression),
           scratch=args.output)



import pyproj

import numpy as np


class Grid():

    def __init__(self, lat, lon, gridsize, proj, fillvalue):

        if (proj == 'conformal') or (proj == 'equal_angle'):
            self.x1, self.y1 = 180., 90.
        elif proj == 'equal_area':
            eqAngle = pyproj.Proj(proj='longlat')
            eqArea = pyproj.Proj(proj='cea')
            self.x1, self.y1 = pyproj.transform(eqAngle, eqArea, 180, 90)
            lon, lat = pyproj.transform(eqAngle, eqArea, lon, lat)
        else:
            raise ValueError('Unkonwn projection "{}". Possible options are ' +
                             '"equal_angle" or "equal_area"'.format(proj))

        self.lat = reformat_vector(lat)
        self.lon = reformat_vector(lon)
        self.gridsize = np.array(gridsize, dtype='f8')
        self.proj = proj
        self.fillvalue = fillvalue

    # function for creating gridded indexes
    def create_index(self):
        """
        """
        if self.proj == 'equal_area':
            gridsize = 2*self.y1/np.floor(2*self.y1/self.gridsize)
        else:
            gridsize = self.gridsize
        # convert coordinates to positive values
        lattemp = self.lat + self.y1
        lontemp = self.lon + self.x1

        # define the sizes of the output matrix depending on the gridsize
        latdim = np.array(np.ceil(2*self.y1/gridsize), dtype='i8')
        londim = np.array(np.ceil(2*self.x1/gridsize), dtype='i8')

        # define the latitude and longitude indexes according to gridsize
        latidx = np.array(np.ceil(lattemp/gridsize) - 1, dtype='i8')
        lonidx = np.array(np.ceil(lontemp/gridsize) - 1, dtype='i8')

        # remove any zero indexes in latidx and lonidx
        idx = np.where(latidx < 0)
        latidx[idx] = 0
        idx = np.where(lonidx < 0)
        lonidx[idx] = 0
        self.edges = [np.min(lonidx), np.max(lonidx),
                      np.min(latidx), np.max(latidx)]

# THIS PART WAS DEFINED FOR THE EQUAL_AREA PROJECTION
# WHENEVER WE BELIEVE WE ARE READY TO MOVE BACK TO IT
# WE CAN POSSIBLY START FROM HERE
        # define the latitude and longitude grids, given the gridsize
        latgrid = np.arange(-self.y1+gridsize/2, self.y1+gridsize-gridsize/2, gridsize)
        longrid = np.arange(-self.x1+gridsize/2, self.x1+gridsize-gridsize/2, gridsize)

        if self.proj == 'equal_area':
            eqArea = pyproj.Proj(proj='cea')
            eqAngle = pyproj.Proj(proj='longlat')
            _, latgrid = pyproj.transform(eqArea, eqAngle,
                                          np.ones((len(latgrid), 1))*longrid[0], latgrid)
            longrid, _ = pyproj.transform(eqArea, eqAngle,
                                          longrid, np.ones((len(longrid), 1))*latgrid[0])

        grididx = np.ravel_multi_index([[lonidx], [latidx]], (londim, latdim))[0]

        grid = {}
        grid['idx'] = grididx
        grid['matrixsize'] = (londim, latdim)
        grid['latgrid'] = latgrid
        grid['longrid'] = longrid

        return grid


class ComputeVariables(Grid):

    def __init__(self, lat, lon, gridsize, proj, fillvalue):
        super().__init__(lat, lon, gridsize, proj, fillvalue)

        self.grid = Grid.create_index(self)
        self.init_size = self.grid['matrixsize'][0] * self.grid['matrixsize'][1]

        # sort indexes
        self.srt_idx = np.sort(self.grid['idx'], kind='mergesort')
        self.sortidx = np.argsort(self.grid['idx'], kind='mergesort')

        # create indexes for the output
        diff_vec = np.diff(self.srt_idx)
        diff_vec = np.insert(diff_vec, 0, 1)
        diff_vec = np.append(diff_vec, 1)
        self.box_idx = np.array(np.where(diff_vec > 0))[0]

#         self.vargrid = {}

    # function for creating gridded variables
    def compute_stats(self, var, varname=''):
        """
        """
        var = reformat_vector(var)

        # check if the sizes of the coordinates passed to the function are consistent
        if np.shape(self.lat) != np.shape(self.lon) or np.shape(self.lat) != np.shape(var):
            raise ValueError('Input coordinates have different size')

        # initialize output variables
        varmean = np.zeros(self.init_size)*np.nan
        varstd = np.zeros(self.init_size)*np.nan
        varsum = np.zeros(self.init_size)*np.nan
        varnpts = np.zeros(self.init_size)*np.nan
        varssum = np.zeros(self.init_size)*np.nan

        # loop for the computation of the output variables
        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            # grab data from var, make sure there are no nans
            xx = var[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            xx = xx[~np.isnan(xx)]
            if (len(xx) > 0):
                varmean[self.srt_idx[self.box_idx[i]]] = np.mean(xx)
                varstd[self.srt_idx[self.box_idx[i]]] = np.std(xx)
                varsum[self.srt_idx[self.box_idx[i]]] = np.sum(xx)
                varnpts[self.srt_idx[self.box_idx[i]]] = len(xx)
                varssum[self.srt_idx[self.box_idx[i]]] = np.sum(xx**2)
            elif (len(xx) == 0):
                varmean[self.srt_idx[self.box_idx[i]]] = np.nan
                varstd[self.srt_idx[self.box_idx[i]]] = np.nan
                varsum[self.srt_idx[self.box_idx[i]]] = np.nan
                varnpts[self.srt_idx[self.box_idx[i]]] = np.nan
                varssum[self.srt_idx[self.box_idx[i]]] = np.nan
            else:
                continue

        # Set all NaN values to fill, see issue-15
        varmean[np.isnan(varmean)] = self.fillvalue
        varstd[np.isnan(varstd)] = self.fillvalue
        varsum[np.isnan(varsum)] = self.fillvalue
        varnpts[np.isnan(varnpts)] = 0  # self.fillvalue
        varssum[np.isnan(varssum)] = self.fillvalue

        if varname != '':
            varname += '/'
        # write the outputs into the output dictionary
        vargrid = {}
        vargrid[varname + 'Mean'] = np.reshape(varmean, self.grid['matrixsize'])
        vargrid[varname + 'Standard_Deviation'] = np.reshape(varstd, self.grid['matrixsize'])
        vargrid[varname + 'Sum'] = np.reshape(varsum, self.grid['matrixsize'])
        vargrid[varname + 'Pixel_Counts'] = np.reshape(varnpts, self.grid['matrixsize'])
        vargrid[varname + 'Sum_Squares'] = np.reshape(varssum, self.grid['matrixsize'])

        return vargrid

    # function for creating gridded variables
    def compute_median(self, var, varname, varbins, resolution):
        """
        """
        var = reformat_vector(var)

        # check if the sizes of the coordinates passed to the function are consistent
        if np.shape(self.lat) != np.shape(self.lon) or np.shape(self.lat) != np.shape(var):
            raise ValueError('Input coordinates have different size')

        # initialize output variables
        varmedian = np.zeros(self.init_size)*np.nan
        median_hist = np.zeros((self.init_size, np.size(varbins)-1))
        # loop for the computation of the output variables
        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            # grab data from var, make sure there are no nans
            xx = var[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            xx = xx[~np.isnan(xx)]
            if (len(xx) > 0):
                median_dict, edges = yori_median(xx, varbins, resolution)
                varmedian[self.srt_idx[self.box_idx[i]]] = median_dict['est_median']
                median_hist[self.srt_idx[self.box_idx[i]], :] = median_dict['hist']
            elif (len(xx) == 0):
                varmedian[self.srt_idx[self.box_idx[i]]] = np.nan
                median_hist[self.srt_idx[self.box_idx[i]], :] = np.nan
            else:
                continue

        # Set all NaN values to fill, see issue-15
        varmedian[np.isnan(varmedian)] = self.fillvalue
        median_hist[np.isnan(median_hist)] = self.fillvalue

        # write the outputs into the output dictionary
        vargrid = {}
        vargrid[varname + '/Median'] = np.reshape(varmedian, self.grid['matrixsize'])
        vargrid[varname + '/Median_Distribution'] = np.reshape(median_hist,
                                                               (self.grid['matrixsize'][0],
                                                                self.grid['matrixsize'][1],
                                                                np.size(varbins)-1))

        return vargrid

    # function for creating weighted stats
    def compute_wgt_stats(self, var, varname='', weight=''):
        """
        """
        if weight == '':
            raise ValueError('Weight not specified when requesting the weighted ' +
                             'stats in the configuration file')

        var = reformat_vector(var)
        weight = reformat_vector(weight)
        # check if the sizes of the coordinates passed to the function are consistent
        if np.shape(self.lat) != np.shape(self.lon) or np.shape(self.lat) != np.shape(var):
            raise ValueError('Input coordinates have different size')

        # initialize output variables
        var_wgt = np.zeros(self.init_size)*np.nan
        var_wgt_sum = np.zeros(self.init_size)*np.nan
        var_wgt_ssum = np.zeros(self.init_size)*np.nan
        var_wgt_mean = np.zeros(self.init_size)*np.nan
        var_wgt_std = np.zeros(self.init_size)*np.nan

        # loop for the computation of the output variables
        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            # grab data from var, make sure there are no nans
            xx = var[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            wgt = weight[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            not_nan = ~np.isnan(xx)
            xx = xx[not_nan]
            wgt = wgt[not_nan]
            mean_wgt = np.sum(xx*wgt)/np.sum(wgt)
            std_wgt = np.sqrt(np.sum(wgt*((xx - mean_wgt)**2))/np.sum(wgt))
            if (len(xx) > 0):
                var_wgt[self.srt_idx[self.box_idx[i]]] = np.sum(wgt)
                var_wgt_sum[self.srt_idx[self.box_idx[i]]] = np.sum(xx*wgt)
                var_wgt_ssum[self.srt_idx[self.box_idx[i]]] = np.sum(wgt*xx**2)
                var_wgt_mean[self.srt_idx[self.box_idx[i]]] = mean_wgt
                var_wgt_std[self.srt_idx[self.box_idx[i]]] = std_wgt
            elif (len(xx) == 0):
                var_wgt[self.srt_idx[self.box_idx[i]]] = np.nan
                var_wgt_sum[self.srt_idx[self.box_idx[i]]] = np.nan
                var_wgt_ssum[self.srt_idx[self.box_idx[i]]] = np.nan
                var_wgt_mean[self.srt_idx[self.box_idx[i]]] = np.nan
                var_wgt_std[self.srt_idx[self.box_idx[i]]] = np.nan
            else:
                continue

        # Set all NaN values to fill, see issue-15
        var_wgt_mean[np.isnan(var_wgt_mean)] = self.fillvalue
        var_wgt_std[np.isnan(var_wgt_std)] = self.fillvalue
        var_wgt_sum[np.isnan(var_wgt_sum)] = self.fillvalue
        var_wgt[np.isnan(var_wgt)] = 0
        var_wgt_ssum[np.isnan(var_wgt_ssum)] = self.fillvalue

        if varname != '':
            varname += '/'
        # write the outputs into the output dictionary
        vargrid = {}
        vargrid[varname + 'Weights'] = np.reshape(var_wgt, self.grid['matrixsize'])
        vargrid[varname + 'Weighted_Sum'] = np.reshape(var_wgt_sum,
                                                       self.grid['matrixsize'])
        vargrid[varname + 'Weighted_Sum_Squares'] = np.reshape(var_wgt_ssum,
                                                               self.grid['matrixsize'])
        vargrid[varname + 'Weighted_Mean'] = np.reshape(var_wgt_mean,
                                                        self.grid['matrixsize'])
        vargrid[varname + 'Weighted_Standard_Deviation'] = np.reshape(var_wgt_std,
                                                                      self.grid['matrixsize'])

        return vargrid

    def compute_histogram(self, var, binvec, varname):
        var = reformat_vector(var)

        varstat = np.zeros((self.init_size, np.size(binvec)-1))  # *np.nan

        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            xx = var[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            xx = xx[~np.isnan(xx)]
            # xx = xx[(~np.isnan(xx)) & (xx != -666) & (xx != -333)]
            if (len(xx) > 0):
                varstat[self.srt_idx[self.box_idx[i]], :] = np.histogram(xx, binvec)[0]
            elif (len(xx) == 0):
                varstat[self.srt_idx[self.box_idx[i]], :] = np.nan
            else:
                continue

        varstat[np.isnan(varstat)] = self.fillvalue
        vargrid = {}
        vargrid[varname + '/Histogram_Counts'] = np.reshape(varstat,
                                                            (self.grid['matrixsize'][0],
                                                             self.grid['matrixsize'][1],
                                                             np.size(binvec)-1))
        return vargrid

    def compute_2D_histogram(self, var1, var2, binvec1, binvec2, varname, var_name_out):
        var1 = reformat_vector(var1)
        var2 = reformat_vector(var2)

        # check if the sizes of the coordinates passed to the function are consistent
        if (np.shape(self.lat) != np.shape(self.lon) or
           np.shape(self.lat) != np.shape(var1) or
           np.shape(self.lat) != np.shape(var2)):
            raise ValueError('Input coordinates have different size')

    # initialize output variables
        var2Dstat = np.zeros((self.init_size, np.size(binvec1)-1, np.size(binvec2)-1))  # *np.nan

    # loop for the computation of the output variables
        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            xx = var1[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            yy = var2[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            nan_idx = np.isnan(xx) | np.isnan(yy)
            xx = xx[~nan_idx]
            yy = yy[~nan_idx]
            # idx_333 = xx == -333 | yy == -333
            # idx_666 = xx == -666 | yy == -666
            # xx = xx[(~nan_idx) & (~idx_333) & (~idx_666)]
            # yy = yy[(~nan_idx) & (~idx_333) & (~idx_666)]

            if (len(xx) > 0) & (len(yy) > 0):
                var2Dstat[self.srt_idx[self.box_idx[i]], :, :] = (
                    np.histogram2d(xx, yy, bins=(binvec1, binvec2))[0])
            elif ((len(xx) == 0) | (len(yy) == 0)):
                var2Dstat[self.srt_idx[self.box_idx[i]], :, :] = np.nan
            else:
                continue

        var2Dstat[np.isnan(var2Dstat)] = self.fillvalue
        vargrid = {}
        vargrid[varname + '/' + var_name_out] = np.reshape(var2Dstat,
                                                           (self.grid['matrixsize'][0],
                                                            self.grid['matrixsize'][1],
                                                            np.size(binvec1)-1,
                                                            np.size(binvec2)-1))

        return vargrid

    def compute_minmax(self, var, varname):
        var = reformat_vector(var)

        # check if the sizes of the coordinates passed to the function are consistent
        if np.shape(self.lat) != np.shape(self.lon) or np.shape(self.lat) != np.shape(var):
            raise ValueError('Input coordinates have different size')

        # initialize output variables
        varmax = np.zeros(self.init_size)*np.nan
        varmin = np.zeros(self.init_size)*np.nan

        # loop for the computation of the output variables
        for i in range(len(self.box_idx)):
            if len(self.box_idx) <= 1:
                continue
            elif i != len(self.box_idx) - 1:
                start = i
                end = i + 1
            else:
                continue

            # grab data from var, make sure there are no nans
            xx = var[self.sortidx[self.box_idx[start]:self.box_idx[end]]]
            xx = xx[~np.isnan(xx)]
            if (len(xx) > 0):
                varmax[self.srt_idx[self.box_idx[i]]] = np.max(xx)
                varmin[self.srt_idx[self.box_idx[i]]] = np.min(xx)
            elif (len(xx) == 0):
                varmax[self.srt_idx[self.box_idx[i]]] = np.nan
                varmin[self.srt_idx[self.box_idx[i]]] = np.nan
            else:
                continue

        # Set all NaN values to fill, see issue-15
        varmax[np.isnan(varmax)] = self.fillvalue
        varmin[np.isnan(varmin)] = self.fillvalue

        if varname != '':
            varname += '/'
        # write the outputs into the output dictionary
        vargrid = {}
        vargrid[varname + 'Max'] = np.reshape(varmax, self.grid['matrixsize'])
        vargrid[varname + 'Min'] = np.reshape(varmin, self.grid['matrixsize'])

        return vargrid


def yori_median(var, binvec, resolution):

    # bins = np.arange(0, max_value+resolution, resolution)
    bins = binvec
    hist, edges = np.histogram(var, bins)
    median_out = estimate_median(hist, edges, resolution)
    return median_out, edges


def estimate_median(hist, edges, resolution):
    msg1 = ('Median Error: something is wrong with i_neg or i_pos that I did not '
            'anticipate. If you see this error, please send an email to '
            'paolo.veglio@ssec.wisc.edu providing a test case able to reproduce the '
            'issue. Thanks.')
    msg2 = ('Median Error: corr_factor should never be negative. If you see this error '
            'send an email to paolo.veglio@ssec.wisc.edu providing a test case able to '
            'reproduce the issue. Thanks.')

    if np.sum(hist) == 0:
        return {'est_median': np.nan,
                'hist': hist}

    rel_area = resolution * (np.cumsum(hist) - np.cumsum(hist)[-1]/2)

    if np.sum(hist) == 1:
        idx = np.nonzero(hist != 0)[0]
        return {'est_median': edges[idx],
                'hist': hist}

    i_zero = np.nonzero(rel_area==0)[0]
    i_pos = np.nonzero(rel_area>0)[0]
    i_neg = np.nonzero(rel_area<0)[0]

    num_count = hist[i_pos[0]]
    corr_factor = -rel_area[i_pos[0]-1]

    if corr_factor == 0 and len(i_neg) > 0:
        est_median = (edges[i_neg[-1]+1] + edges[i_pos[0]])/2
    elif len(i_neg) == 0:
        if len(i_zero) > 0:
            est_median = edges[i_zero[0]]
        elif len(i_pos) > 0:
            est_median = edges[i_pos[0]]
        else:
            # this should never happen. I'm leaving it for now in case something weird
            # makes the code enter this condition
            raise ValueError(msg1)
    elif corr_factor > 0:
        est_median = edges[i_pos[0]] + corr_factor/num_count
    else:
        # this should never happen. I'm leaving it for now in case something weird
        # makes the code enter this condition
        raise ValueError(msg2)

    median_out = {'est_median': est_median,
                  'hist': hist}

    return median_out


def reformat_vector(vec):
    if vec.ndim > 1:
        vec = np.squeeze(vec)
        vec = np.reshape(vec, np.shape(vec)[0]*np.shape(vec)[1])

    vec = np.array(vec, dtype='f8')
    return vec



import sys
import warnings

import netCDF4 as nc
import numpy as np

import yori.IOtools as IO
import yori.config_reader as cfg
import yori.aggrtools as tools

# from datetime import datetime
from pkg_resources import get_distribution


NC_FILL_DOUBLE = -9999.
VERSION = get_distribution('yori').version


########################################
#               FUNCTION               #
########################################
#
def aggregate(flist, fname_out, satellite='', daily='',
              compression=5, verbose=False, force=False,
              use_min_pixel_counts=False, use_min_valid_days=False,
              batch_size=2):

    warnings.simplefilter('module')

    f0 = nc.Dataset(flist[0])
    group_list = list(f0.groups)

    latdim = f0.dimensions['latitude'].size
    londim = f0.dimensions['longitude'].size

    grid_vers = f0.getncattr('Yori_version')
    if grid_vers != str(VERSION):
        message = ("Current yori-aggr version (v" + str(grid_vers) +
                   ") doesn't match the version used to create the gridded " +
                   "files that are being aggregated")
        warnings.warn(message, DeprecationWarning)

    config_str = f0.getncattr('YAML_config')
    config_str = str(config_str[:])
    ymlSet = cfg.ConfigReader(config_str)

    latkey = 'latitude'
    lonkey = 'longitude'

    gridsize = ymlSet.grid_settings['gridsize']

    if 'fill_value' in ymlSet.grid_settings:
        fv = ymlSet.grid_settings['fill_value']
    else:
        fv = NC_FILL_DOUBLE

    lat = f0[latkey][:]
    lon = f0[lonkey][:]

    for f in flist:
        fin = nc.Dataset(f)
        if hasattr(fin, 'daily_defn_of_day_adjustment'):
            if fin.getncattr('daily_defn_of_day_adjustment') == 'True':
                message = ('The input files have been created with the ' +
                           '--daily flag active. This could lead to small ' +
                           'differences in the creation of the aggregated ' +
                           'files when compared to monthly products created ' +
                           'without using this flag. See documentation for ' +
                           'more details')
                warnings.warn(message)

        fin.close()

    if daily != '':
        dmask = tools.daily_mask(satellite, daily, flist, gridsize)
    else:
        dmask = np.ones((londim, latdim))

    f0.close()

    # initialize output file
    nc0 = nc.Dataset(fname_out, mode='w', format='NETCDF4')
    nc0.createDimension('latitude', len(lat))
    nc0.createDimension('longitude', len(lon))
    nc0.createVariable('latitude', 'f8', ('latitude', ), fill_value=fv)
    nc0.createVariable('longitude', 'f8', ('longitude', ), fill_value=fv)
    nc0['latitude'][:] = lat
    nc0['longitude'][:] = lon

    # initialize flags that will be used to avoid repeating warnings at each iteration
    warned_min_pixel = False
    warned_min_days = False

    # aggregate the variable groups in batches to keep memory footprint down
    batching = batch_size

    for batch_idx in range(0, len(group_list), batching):
        group_batch = group_list[batch_idx:batch_idx + batching]
        if verbose:
            print('processing: ' + ', '.join(group_batch))

        # process all the files for the current batch
        arrays_initialized = False
        for fin in flist:

            file_in = nc.Dataset(fin)

            if not arrays_initialized:
                var_in = tools.initialize_arrays(group_batch, file_in, latkey, lonkey, fv)
                arrays_initialized = True

            # look for global attribute that describes the subset of the full
            # global grid that actually contains data; enables working with much
            # smaller arrays even when Pixel_Counts are not present
            granule_edges = getattr(file_in, 'granule_edges', None)
            if granule_edges is not None:
                left, right, bottom, top = granule_edges
                x_slice = slice(left, right + 1)
                y_slice = slice(bottom, top + 1)
                granule_edges = x_slice, y_slice

            if daily != '':
                try:
                    timeattr = file_in.getncattr('time_coverage_start')
                except AttributeError:
                    raise AttributeError('Global attribute "time_coverage_start"' +
                                         'not found in input file ' + fin)
                if satellite == 'aqua':
                    dmask = np.asarray(tools.aqua_daily_aggr(timeattr, daily, gridsize),
                                       bool)
                if satellite == 'terra':
                    dmask = np.asarray(tools.terra_daily_aggr(timeattr, daily, gridsize),
                                       bool)
            else:
                dmask = np.ones((londim, latdim), bool)

            # initialize the days and months arrays that are used to check whether the
            # files passed to aggregate belong to the same day/month. This is useful
            # when the min_pixel_counts and min_valid_days are used.
            days, months = [], []

            # Note: the reason to append a random value to days/months if the attribute
            # 'time_coverage_start' is not defined is to make sure that the following
            # check (every file belongs to same day/month) throws a warning if this one
            # throws a warning
            if use_min_pixel_counts is True:
                try:
                    timeattr = file_in.getncattr('time_coverage_start')
                    days.append(timeattr.day)
                except AttributeError:
                    message = (('Global attribute "time_coverage_start" not found in ' +
                                'input file {0}. Yori is unable to verify if the input ' +
                                'files are all from the same day. Please double check ' +
                                'your inputs. This message will not be repeated for ' +
                                'other files.').format(fin))
                    if warned_min_pixel is False:
                        warnings.warn(message)
                        warned_min_pixel = True
                    days.append(np.random.randint(-100000, -1))

            if use_min_valid_days is True:
                try:
                    timeattr = file_in.getncattr('time_coverage_start')
                    months.append(timeattr.month)
                except AttributeError:
                    message = (('Global attribute "time_coverage_start" not found in ' +
                                'input file {0}. Yori is unable to verify if the input ' +
                                'files are all from the same month. Please double ' +
                                'check your inputs. This message will not be repeated ' +
                                'for other files').format(fin))
                    if warned_min_days is False:
                        warnings.warn(message)
                        warned_min_days = True
                    months.append(np.random.randint(-100000, -1))

            var_add = {}
            var_add = tools.prepare_vars(var_add, group_batch, file_in, granule_edges)
            for k in group_batch:
                if k == latkey or k == lonkey or k not in var_add:
                    pass
                else:
                    # use optimized aggregation function that avoids large array
                    # copies
                    var_in[k] = tools.gq_aggregate(var_in[k], var_add[k], dmask, fv)
            file_in.close()

        # check that all the files belong to the same day or month when the
        # min_pixel_counts or min_valid_days flags are used
        if use_min_pixel_counts is True:
            if len(set(days)) != 1:
                message = ('Warning: not all files belong to the same day. Please ' +
                           'verify your inputs. If this is intended please disregard ' +
                           'this message')
                warnings.warn(message)
        if use_min_valid_days is True:
            if len(set(months)) != 1:
                message = ('Warning: not all files belong to the same month. Please ' +
                           'verify your inputs. If this is intended please disregard ' +
                           'this message')
                warnings.warn(message)

        # replace fill and compute mean and stddev now, instead of each
        # time through the loop
        # var_in gets updated with the newly computed and modified values
        tools.compute_aggr_stats(var_in, fv)

        # filter the output variables based on the minimum number of valid pixels allowed
        # by the user. This option should only be used when creating D3 files. A warning
        # is thrown if the dates don't belong to the same day
        if use_min_pixel_counts is True:
            for grp in var_in.values():
                if 'Pixel_Counts' not in grp:
                    continue
                if 'min_pixel_counts' in grp:
                    min_pixel_counts = grp['min_pixel_counts']
                    for k in grp.keys():
                        if k == 'min_pixel_counts':
                            continue
                        grp['Pixel_Counts'][grp['Pixel_Counts'] < min_pixel_counts] = 0
                        if k not in ['Pixel_Counts', 'min_valid_days']:
                            grp[k][grp['Pixel_Counts'] < min_pixel_counts] = fv

                    grp.pop('min_pixel_counts')

        for grp in var_in.values():
            if 'min_pixel_counts' in list(grp):
                grp.pop('min_pixel_counts')
            if 'min_valid_days' in list(grp):
                grp.pop('min_valid_days')
            if 'Num_Days' in list(grp) and use_min_valid_days is False:
                grp.pop('Num_Days')
            if 'edges' in list(grp):
                grp.pop('edges')

        var_in = tools.restructDict(var_in)

        grid_size = ymlSet.grid_settings['gridsize']
        IO.write_output(nc0, var_in, compression, grid_size, ymlSet, fv)

    # write the config file into a variable inside the netcdf output file
    nc0.setncattr('YAML_config', ymlSet.yaml_str)
    nc0.setncattr('Yori_version', str(VERSION))
    nc0.setncattr('input_files', ','.join(flist))
    nc0.setncattr('daily_defn_of_day_adjustment', str(False))
    if daily != '':
        nc0.setncattr('daily_defn_of_day_adjustment', str(True))

    for i in range(len(ymlSet.var_settings)):
        settings = ymlSet.var_settings[i]
        IO.append_attributes(nc0, settings)

    setattr(nc0.variables[latkey], 'units', 'degrees_north')
    setattr(nc0.variables[lonkey], 'units', 'degrees_east')

    nc0.close()


# main
if __name__ == "__main__":

    if len(sys.argv) == 1:
        print('usage: yori_aggregate.py <file_list> <output_file>')
        sys.exit(1)

    file_list = sys.argv[1]
    fname_out = sys.argv[2]

    aggregate(file_list, fname_out)



import sys
import os.path
import warnings

import numpy as np
import netCDF4 as nc

from pkg_resources import get_distribution

import yori.gridtools as grid
import yori.config_reader as cfg
import yori.IOtools as IO

VERSION = get_distribution('yori').version
_NC_FILL_DOUBLE = -9999.
# temp workaround, not sure if there are better ways to make yori packages
# work from shell probably I need to add the entire path to PYTHONPATH
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))


class SetupYori():

    def __init__(self, config_file, input_file, output_file):
        self.input_file = input_file
        self.output_file = output_file
        self.outpath, self.outfn = os.path.split(output_file)
        warnings.simplefilter('module')

        if os.path.isfile(output_file):
            raise IOError('The file {outfn} already exists in ' +
                          'the destination folder'.format(outfn=self.outfn))

        if config_file.lower().endswith('.csv'):
            cfg.write_config(config_file)
            config_file = config_file[:-3] + 'yml'

        cf = open(config_file)
        config_str = cf.read()
        cf.close()
        self.ymlSet = cfg.ConfigReader(config_str)
        self.grid_proj = self.ymlSet.grid_settings['projection']
        self.grid_size = np.array(self.ymlSet.grid_settings['gridsize'],
                                  dtype='f8')
        if self.grid_proj == 'equal_area':
            self.grid_size = 1000*self.grid_size

        if ('lat_in' not in self.ymlSet.grid_settings or
           'lon_in' not in self.ymlSet.grid_settings):

            warnings.warn('Coordinate names are not defined in the configuration file. '
                          'This format will be discontinued in future versions',
                          DeprecationWarning)

            self.lat_key = 'latitude'
            self.lon_key = 'longitude'

        else:
            self.lat_key = self.ymlSet.grid_settings['lat_in']
            self.lon_key = self.ymlSet.grid_settings['lon_in']

        if 'fill_value' in self.ymlSet.grid_settings:
            self.fv = self.ymlSet.grid_settings['fill_value']
        else:
            self.fv = _NC_FILL_DOUBLE

    def setCoordinates(self):

        tmplat = np.zeros((1,))
        tmplon = np.zeros((1,))
        # NOTE:
        # this part works as long as the projection is equal angle.
        # Once we introduce different projections this part needs to be changed
        myCoords = grid.Grid(tmplat, tmplon, self.grid_size, self.grid_proj, self.fv)
        gridvar = {}
        coord_grid = myCoords.create_index()
        gridvar['latitude'] = coord_grid['latgrid']
        gridvar['longitude'] = coord_grid['longrid']
#        gridvar['latitude'] = np.arange(-90.+self.grid_size/2.,
#                                        90.+self.grid_size-self.grid_size/2.,
#                                        self.grid_size)
#        gridvar['longitude'] = np.arange(-180.+self.grid_size/2.,
#                                         180.+self.grid_size-self.grid_size/2.,
#                                         self.grid_size)

        return gridvar

    def setVariable(self, var_settings, lat, data_in, joint=False):
        if joint is False:
            var_name_in = var_settings['name_in']
        else:
            var_name_in = var_settings['joint_var']['name_in']

        var_name_out = var_settings['name_out']

        mask_list, inv_mask_list, extra_mask_list = [], [], []
        if 'masks' in var_settings:
            mask_list = var_settings['masks']
        if 'inverse_masks' in var_settings:
            inv_mask_list = var_settings['inverse_masks']
        if 'extra_masks' in var_settings:
            extra_mask_list = var_settings['extra_masks']

        current_mask = create_mask(data_in, np.shape(lat), mask_list=mask_list,
                                   inv_mask_list=inv_mask_list)
        current_mask = create_mask(data_in, np.shape(lat),
                                   mask_list=extra_mask_list,
                                   existing_mask=current_mask)

        mask_idx = np.nonzero(current_mask == 0)
        tmp_data_in = data_in[var_name_in][:]
        if hasattr(tmp_data_in, 'filled'):
            tmp_data_in = tmp_data_in.filled(np.nan)
        masked_data_in = tmp_data_in
        masked_data_in[mask_idx] = np.nan
        return masked_data_in, var_name_out

    def runYori(self, debug=False, compression=0):

        data_in = nc.Dataset(self.input_file)
        gridvar = self.setCoordinates()

        lat = data_in[self.lat_key][:]
        lon = data_in[self.lon_key][:]
        myGrid = grid.ComputeVariables(lat, lon, self.grid_size,
                                       self.grid_proj, self.fv)

        data_out = nc.Dataset(self.output_file, 'w')

        data_out.setncattr('YAML_config', self.ymlSet.yaml_str)
        data_out.setncattr('Yori_version', str(VERSION))

        for i in range(len(self.ymlSet.var_settings)):
            var_settings = self.ymlSet.var_settings[i]

            if 'only_histograms' not in var_settings:
                only_hist = False
            else:
                only_hist = True

            # 'only_weights' is temporary, when defined in cfg, change accordingly
            if 'only_weights' not in var_settings:
                only_wgt = False
            else:
                only_wgt = True

            masked_data_in, var_name_out = self.setVariable(var_settings,
                                                            lat, data_in)

            if only_hist is False and only_wgt is False:
                tmp_gridvar = myGrid.compute_stats(masked_data_in, var_name_out)
                gridvar.update(tmp_gridvar)

            if 'median' in var_settings:
                varbins = self.ymlSet.readHist(var_settings['median'])
                tmp_gridvar = myGrid.compute_median(masked_data_in, var_name_out,
                                                    varbins,
                                                    var_settings['median']['interval'])
                gridvar.update(tmp_gridvar)

            if 'minmax' in var_settings:
                tmp_gridvar = myGrid.compute_minmax(masked_data_in, var_name_out)
                gridvar.update(tmp_gridvar)

            if 'weights' in var_settings:
                wgt = data_in[var_settings['weights']]
                tmp_gridvar = myGrid.compute_wgt_stats(masked_data_in, var_name_out,
                                                       weight=wgt)
                gridvar.update(tmp_gridvar)

            if 'histograms' in var_settings:
                varbins = self.ymlSet.readHist(var_settings['histograms'])
                tmp_gridvar = myGrid.compute_histogram(masked_data_in, varbins,
                                                       var_name_out)
                gridvar.update(tmp_gridvar)

            if '2D_histograms' in var_settings:
                for k in range(len(var_settings['2D_histograms'])):

                    hist_2D_settings = var_settings['2D_histograms'][k]
                    varbins = self.ymlSet.readHist(hist_2D_settings['primary_var'])
                    joint_varbins = self.ymlSet.readHist(hist_2D_settings['joint_var'])

                    joint_masked_data_in, vno = self.setVariable(hist_2D_settings, lat,
                                                                 data_in, joint=True)

                    joint_var_name_out = hist_2D_settings['name_out']
                    tmp_gridvar = myGrid.compute_2D_histogram(masked_data_in,
                                                              joint_masked_data_in,
                                                              varbins, joint_varbins,
                                                              var_name_out,
                                                              joint_var_name_out)
                    gridvar.update(tmp_gridvar)

            IO.write_output(data_out, gridvar, compression, self.grid_size,
                            self.ymlSet, self.fv,
                            edges=myGrid.edges, write_edges=True)
            IO.append_attributes(data_out, var_settings)
            gridvar = {}

        data_in.close()
        data_out.close()


def callYori(config_file, input_file, output_file, debug=False, compression=0):

    Yori = SetupYori(config_file, input_file, output_file)
    Yori.runYori(debug, compression)


def create_mask(in_data, init_size, mask_list=[], inv_mask_list=[],
                existing_mask=None):
    if existing_mask is None:
        out_mask = np.array(np.ones(init_size), dtype='i8')
    else:
        out_mask = existing_mask

    for mask in mask_list:
        update_mask = np.array(in_data[mask][:], dtype='i8')
        out_mask &= update_mask

    # Note: this doesn't have to be used for inverse masks. Another possible
    # use is to update existing masks, like the case of 2D histograms with
    # extra masks that are being applied on top of any existing mask
    for inverse_mask in inv_mask_list:
        tmp_inv_mask = np.array(in_data[inverse_mask][:], dtype='i8')
        update_mask = np.abs(tmp_inv_mask - 1)
        out_mask &= update_mask

    return out_mask


# main
if __name__ == "__main__":

    if len(sys.argv) == 1:
        print('usage: rungrid.py <config_file> <input_file> <output_file>')
        sys.exit(1)

    config_file = sys.argv[1]
    input_file = sys.argv[2]
    output_file = sys.argv[3]
    comp = sys.argv[4] if len(sys.argv) == 5 else 0

    callYori(config_file, input_file, output_file, compression=comp)



from pkg_resources import get_distribution
from ..cluster_tools import process_day

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Command that downloads MODIS-CALIOP ' +
                                     'matchfiles, convert them to netcdf4 and grid them')
    parser.add_argument('year', help='Year of the data to be processed')
    parser.add_argument('jday', help='Julian day of the data to be processed')
    parser.add_argument('-V', '--verbose', action='store_true', help='Enable verbose mode.')
    parser.add_argument('-v', '--version', action='version', version=VERSION, help='Show ' +
                        'version and exit')
    parser.add_argument('-j', '--maxjobs', help='number of simultaneous jobs to run')
    parser.add_argument('-d', '--download', action='store_true',
                        help='tell the code to download the files from the ftp server')
    parser.add_argument('-p', '--path',
                        help='local path where the data will be saved and processed')

    args = parser.parse_args()

    process_day(args.year,
                args.jday,
                verbose=args.verbose,
                maxjobs=args.maxjobs,
                ftp_download=args.download,
                local_path=args.path)



import os
import subprocess
import ftplib
import numpy as np
import sys
# import netCDF4 as nc

from copy import deepcopy

import yori.ioutils as io

from utils.readers import read_hdf

# _ftp_path = 'sips.ssec.wisc.edu/products/matchups/modis-caliop-alexa/v20170801/'
_ftp_path = 'sips.ssec.wisc.edu/match/modis-caliop/1.1dev6/'
_iris_path = '/data/pveglio/NRL-NAV/'
# _var_list = ['modis_latitude', 'modis_longitude', 'modis_landsea_mask',
#              'modis_daynight_mask', 'modis_cloud_top_height', 'modis_cloud_mask',
#              'modis_cloud_optical_thickness', 'modis_aod_unfiltered',
#              'caliop_clay_time', 'caliop_clay_layer_top_altitude',
#              'caliop_clay_cloud_fraction', 'caliop_cpro_cloud_extinction_msl',
#              'caliop_apro_aerosol_extinction_msl', 'caliop_clay_opacity_flag',
#              'modis_cloud_phase']

_var_list = ['master_latitude', 'master_longitude', 'master_landsea_mask',
             'master_daynight_mask', 'master_cloud_top_height', 'master_cloud_mask_myd35',
             'master_cloud_optical_thickness', 'master_aod_unfiltered',
             'caliop_clay_time', 'caliop_clay_layer_top_altitude',
             'caliop_clay_cloud_fraction', 'caliop_cpro_cloud_extinction_msl',
             'caliop_apro_aerosol_extinction_msl', 'caliop_clay_opacity_flag',
             'master_cloud_phase', 'master_solar_zenith', 'caliop_clay_ice_fraction',
             'caliop_clay_liquid_fraction', 'caliop_clay_feature_optical_depth_532',
             'master_cloud_phase', 'caliop_clay_cloud_fraction_1km']


def hdf2nc(fname, var_list=_var_list, replace=False, preprocessing=False,
           compression=0, scratch=''):

    if scratch == '':
        fout = fname[:-3] + 'nc'
    else:
        fout = scratch + fname.split('/')[-1][:-3] + 'nc'

    modcal = read_hdf(fname, var_list)
    tmp_modcal = {k: np.squeeze(modcal[k]) for k in list(modcal)}

    if preprocessing is True:
        new_modcal = preproc(tmp_modcal)
    else:
        new_modcal = tmp_modcal

    outInst = io.CreateMatchfile(fout, new_modcal, comp=compression)
    outInst.createFile()

    for k in list(new_modcal):
        outInst.saveNewVar(k)

    if replace is True:
        os.system('rm -rf ' + fname)


def preproc(dict_var):

    ot, tmp, out_dict = {}, {}, {}
    ot['cal_cloud'] = np.nansum(dict_var['caliop_cpro_cloud_extinction_msl'], axis=0)
    opacity_flag = np.nansum(dict_var['caliop_clay_opacity_flag'], axis=0)
    idx = np.nonzero(opacity_flag > 0)
    ot['cal_cloud'][idx] = 100
    ot['cal_aerosol'] = np.nansum(dict_var['caliop_apro_aerosol_extinction_msl'], axis=0)
    ot['cal_total'] = ot['cal_cloud'] + ot['cal_aerosol']

###

    ot['cal_cloud_532'] = np.nansum(dict_var['caliop_clay_feature_optical_depth_532'], axis=0)
    ot['cal_cloud_532'][ot['cal_cloud_532'] < 0] = 0
    tmp['caliop_cf_1km'] = np.ceil(dict_var['caliop_clay_cloud_fraction_1km'][0, :])
    cf_ice = np.nansum(dict_var['caliop_clay_ice_fraction'], axis=0)
    cf_wtr = np.nansum(dict_var['caliop_clay_liquid_fraction'], axis=0)
    cf_tot = cf_ice + cf_wtr
    idx = np.nonzero(cf_tot == 0)
    cf_ice[idx], cf_wtr[idx], cf_tot[idx] = 0., 0., 1.

    ice_fraction = np.floor(cf_ice) / cf_tot
#    water_fraction = cf_wtr / cf_tot

    idx1 = np.nonzero(ice_fraction == 1)
#    idx2 = np.nonzero(water_fraction == 1)
#    idx3 = np.nonzero((ice_fraction < 1) & (water_fraction < 1))
#    idx4 = np.nonzero((ice_fraction == 0) & (water_fraction == 0))

    tmp_ice_mask = np.zeros(np.shape(cf_tot))
    tmp_ice_mask[idx1] = 1

# PLEASE DOUBLE CHECK THESE LINES AND MAKE SURE THEY MAKE SENSE!!!
    tmp['caliop_cf_ice'] = np.zeros(np.shape(tmp['caliop_cf_1km']))
    tmp['caliop_cf_ice'] = ice_fraction

#    tmp['caliop_cf_mixed'] = np.zeros(np.shape(cf_tot))
#    tmp['caliop_cf_mixed'][idx2] = dict_var['caliop_clay_cloud_fraction_1km'][0, idx2]
#    tmp['caliop_cf_water'] = np.zeros(np.shape(cf_tot))
#    tmp['caliop_cf_water'][idx3] = dict_var['caliop_clay_cloud_fraction_1km'][0, idx3]
#    tmp['caliop_cf_clear'] = np.zeros(np.shape(cf_tot))
#    tmp['caliop_cf_clear'][idx4] = dict_var['caliop_clay_cloud_fraction_1km'][0, idx4]

###

    ot['mod_cloud'] = np.squeeze(dict_var['master_cloud_optical_thickness']*0.01)
    ot['mod_aerosol'] = dict_var['master_aod_unfiltered']

    masks = cloud_phase_masks(dict_var['master_cloud_phase'])
    masks['cal_mask_ice'] = tmp_ice_mask

    ot['mod_cloud'][ot['mod_cloud'] == 0] = np.nan
    ot['mod_cloud'][masks['clear'] == 1] = 0
    ot['mod_total'] = np.nansum(np.dstack((ot['mod_cloud'], ot['mod_aerosol'])), 2)
    ot['mod_total'] = np.squeeze(ot['mod_total'])

    tmp['caliop_cloud_fraction'] = dict_var['caliop_clay_cloud_fraction'][0, :]
    tmp['modis_cloud_fraction'] = np.abs(np.floor(dict_var['master_cloud_mask_myd35']/2)-1)
    tmp['modis_cloud_fraction'] = np.squeeze(tmp['modis_cloud_fraction'])

    for k in list(ot):
        tmp[k + '_transmittance'] = compute_transmittance(ot[k])

    transbins = np.concatenate(([100, 98, 95], np.arange(90, 0, -10), [7, 5]))

    for tau in transbins:
        cal_tr = deepcopy(tmp['cal_cloud_transmittance'])
        mod_tr = deepcopy(tmp['mod_cloud_transmittance'])
        cal_tr[np.isnan(cal_tr)] = -1
        mod_tr[np.isnan(mod_tr)] = -1
        cf_mask_cal = deepcopy(tmp['caliop_cloud_fraction'])
        cf_mask_cal[cal_tr > tau/100.] = 0
        tmp['cal_mask_lt_' + str(tau)] = cf_mask_cal

        cf_mask_mod = deepcopy(tmp['modis_cloud_fraction'])
        cf_mask_mod[mod_tr > tau/100.] = 0
        tmp['mod_mask_lt_' + str(tau)] = cf_mask_mod

    out_dict['modis_latitude'] = dict_var['master_latitude']
    out_dict['modis_longitude'] = dict_var['master_longitude']
    out_dict.update(tmp)
    out_dict.update(ot)

    idx_day = np.nonzero(np.squeeze(dict_var['master_solar_zenith']) < 90.)
    idx_night = np.nonzero(np.squeeze(dict_var['master_solar_zenith']) >= 90.)
    out_dict['modis_day_mask'] = np.zeros(np.shape(np.squeeze(dict_var['master_solar_zenith'])))
    out_dict['modis_night_mask'] = np.zeros(np.shape(np.squeeze(dict_var['master_solar_zenith'])))
    out_dict['modis_day_mask'][idx_day] = 1
    out_dict['modis_night_mask'][idx_night] = 1

    for k in list(masks):
        out_dict['cloud_mask_' + k] = masks[k]

    return out_dict


def cloud_phase_masks(cpop):
    mask = {}

    mask_list = ['undetermined', 'clear', 'water', 'ice', 'undetermined_phase']
    for k in list(mask_list):
        mask[k] = np.zeros(np.shape(cpop))

    mask['undetermined'][cpop == 0] = 1
    mask['clear'][cpop == 1] = 1
    mask['water'][cpop == 2] = 1
    mask['ice'][cpop == 3] = 1
    mask['undetermined_phase'][cpop == 4] = 1

    return mask


def compute_transmittance(optical_thickness):
    trans = np.exp(-optical_thickness)
    return trans


def download_only(year, local_path=_iris_path, start_day=1, end_day=366):

    days = range(start_day, end_day+1)
    ftp_jobid = 0
    for jday in days:
        grid = SelectDate(year, jday, iris_path=local_path)
        file_list = grid.check_path()
        if len(file_list) > 0:
            grid.set_folders()

            fid = open(grid.wdir + 'flist', 'w')
            for fname in file_list:
                fid.write(fname + '\n')
            fid.close()

            if ftp_jobid == 0:
                ftp_cmd = ('sbatch /home/pveglio/sbatch_scripts/yori_download.sh ' +
                           str(year) + ' ' + format(jday, '03d'))
                ftp_status = subprocess.getstatusoutput(ftp_cmd)
                ftp_jobid = ftp_status[1].strip('Submitted batch job ')
            else:
                ftp_cmd = ('sbatch --dependency=afterany:{0} ' +
                           '/home/pveglio/sbatch_scripts/yori_download.sh ' +
                           str(year) + ' ' + format(jday, '03d')).format(ftp_jobid)
                ftp_status = subprocess.getstatusoutput(ftp_cmd)
                ftp_jobid = ftp_status[1].strip('Submitted batch job ')
            grid_cmd = ('sbatch ' + '--dependency=afterok:{0} ' +
                        '/home/pveglio/sbatch_scripts/sbatch_repack.sh ' +
                        grid.wdir).format(ftp_jobid)
            # print(grid_cmd)
            subprocess.call(grid_cmd, shell=True)


def grid_only(year, local_path=_iris_path, start_day=1, end_day=366):

    days = range(start_day, end_day+1)

    for jday in days:
        grid = SelectDate(year, jday, iris_path=local_path)
        grid_cmd = ('sbatch ' + '/home/pveglio/sbatch_scripts/yori_cluster.sh ' + grid.wdir)
        subprocess.call(grid_cmd, shell=True)


def run_year(year, local_path=_iris_path, start_day=1, end_day=366,
             nfiles=30, ftp_jobid=0, path_slurm_scripts='/home/pveglio/sbatch_scripts/'):

    days = range(start_day, end_day+1)
    for jday in days:
        grid = SelectDate(year, jday, iris_path=local_path)
        file_list = grid.check_path()
        if len(file_list) > 0:
            grid.set_folders()

            fid = open(grid.wdir + 'flist', 'w')
            for fname in file_list:
                fid.write(fname + '\n')
            fid.close()

            n_parts = len(file_list) // nfiles
            partial_flist = split_list(file_list, n_parts)
            save_partial_list(partial_flist, grid.wdir)

            if ftp_jobid == 0:
                ftp_cmd = ('sbatch ' + path_slurm_scripts + 'yori_download.sh ' +
                           str(year) + ' ' + format(jday, '03d'))
                ftp_status = subprocess.getstatusoutput(ftp_cmd)
                ftp_jobid = ftp_status[1].strip('Submitted batch job ')
                # print(ftp_cmd)
            else:
                ftp_cmd = ('sbatch --dependency=afterany:{0} ' +
                           path_slurm_scripts + 'yori_download.sh ' +
                           str(year) + ' ' + format(jday, '03d')).format(ftp_jobid)
                ftp_status = subprocess.getstatusoutput(ftp_cmd)
                ftp_jobid = ftp_status[1].strip('Submitted batch job ')
                # print(ftp_cmd)
            grid_cmd = ('sbatch ' + '--dependency=afterok:{0} ' +
                        '--array=1-{1} ' +
                        path_slurm_scripts + 'yori_cluster.sh ' +
                        grid.wdir).format(ftp_jobid, n_parts)
            # print(grid_cmd)
            subprocess.call(grid_cmd, shell=True)


def split_list(inlist, wanted_parts):
    length = len(inlist)

    return [inlist[i*length // wanted_parts: (i+1)*length // wanted_parts]
            for i in range(wanted_parts)]


def save_partial_list(inlist, path):
    for i in range(len(inlist)):
        f = open('{0}tmp_flist_{1}'.format(path, i), 'w')
        for l in inlist[i]:
            f.write(l + '\n')
        f.close()


class SelectDate(object):

    def __init__(self, year, jday, verbose=False, ftp_path=_ftp_path, iris_path=_iris_path):
        self.year = int(year)
        self.jday = int(jday)
        self.verbose = verbose
        self.ftp_path = ftp_path
        self.iris_path = iris_path
        self.sub_path = str(self.year) + '/' + format(self.jday, '03d') + '/'
        self.wdir = self.iris_path + self.sub_path

    def set_folders(self):
        os.system('mkdir -p ' + self.wdir)

    def check_path(self):
        ftp_root = self.ftp_path.split('/')[0]
        subdir = self.ftp_path.strip(ftp_root)
        ftp = ftplib.FTP(ftp_root, 'anonymous')
        flist = []

        if subdir + str(self.year) in ftp.nlst(subdir):
            ftp.cwd(subdir + str(self.year))

            if format(self.jday, '03d') in ftp.nlst():
                ftp.cwd(format(self.jday, '03d'))
                flist = ftp.nlst('*.hdf')
        ftp.quit()

        return flist


if __name__ == '__main__':

    hdf2nc(fname=sys.argv[1],
           replace=True,
           preprocessing=True)



from setuptools import setup, find_packages

setup(name='yori',
      description='Set of tools to grid satellite data',
      author='Paolo Veglio',
      author_email='paolo.veglio@ssec.wisc.edu',
      url='https://gitlab.ssec.wisc.edu/pveglio/yori',
      use_scm_version=True,
      packages=find_packages(),
      python_requires='>=3.9',
      setup_requires=['setuptools_scm'],
      # install_requires=['numpy', 'ruamel.yaml', 'netcdf4', 'pyproj'],
      entry_points={'console_scripts': [
            'yori-grid = yori.tools.grid:main',
            'yori-aggr = yori.tools.aggr:main',
            'yori-merge = yori.tools.merge:main', ]},
      )



from ..run_yori import callYori
from pkg_resources import get_distribution
# from setuptools_scm import get_version

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Command for gridding geophysical data')
# replace debug with verbose in a future version and have a command to print
# some info when the proper flag is used
#    parser.add_argument('--debug', action='store_true',
#                        help='enable debug; for development only!')
    parser.add_argument('-v', '--version', action='version', version=VERSION,
                        help='show version and exit')
    parser.add_argument('cfgfile', help='Yori yaml config file')
    parser.add_argument('input', help='name and path of the input file in netcdf4 format')
    parser.add_argument('output', help='name and path of the output gridded file')
    parser.add_argument('-c', '--compression', default=5,
                        help='set the compression level for netcdf4 (0=none 9=max)')

    args = parser.parse_args()

    callYori(args.cfgfile,
             args.input,
             args.output,
             compression=args.compression)



from pkg_resources import get_distribution
from ..cluster_tools import run_year

VERSION = get_distribution('yori').version


def main():
    import argparse
    parser = argparse.ArgumentParser(description='Parallelizable command to run yori-grid on ' +
                                     'CALIOP-MODIS matchfiles. It retrieves ' +
                                     'the files from the SIPS ftp server, converts them and ' +
                                     'process them with yori-grid')
    parser.add_argument('-y', '--year', help='Year of the data to be processed')
    parser.add_argument('-p', '--path', default='/iliad/lenny/pveglio/NRL-NAV/',
                        help='Path where the files are going to be saved')
    parser.add_argument('--start', default=1, help='starting day of processing')
    parser.add_argument('--end', default=365, help='ending day of processing')
    parser.add_argument('-n', '--nfiles', default=15,
                        help='Number of files to be processed per job')
    parser.add_argument('--slurm', default='/home/pveglio/sbatch/scripts/',
                        help='Path where the sbatch scripts are saved')
    parser.add_argument('-d', '--dependency', default=0,
                        help='JobID of the job the new batch submission needs to depend on')
    parser.add_argument('-v', '--version', action='version', version=VERSION, help='Show ' +
                        'version and exit')

    args = parser.parse_args()

    run_year(args.year,
             args.path,
             args.start,
             args.end,
             args.nfiles,
             args.dependency,
             args.slurm)
